
//  (C) Copyright Edward Diener 2011,2012
//  Use, modification and distribution are subject to the Boost Software License,
//  Version 1.0. (See accompanying file LICENSE_1_0.txt or copy at
//  http://www.boost.org/LICENSE_1_0.txt).

#if !defined(BOOST_TTI_DETAIL_NOTYPE_HPP)
#define BOOST_TTI_DETAIL_NOTYPE_HPP

namespace boost
  {
  namespace tti
    {
    namespace detail
      {
      struct notype
        {
        };
      }
    }
  }
  
#endif // BOOST_TTI_DETAIL_NOTYPE_HPP
